@@ -10,13 +10,20 @@ CREATE FUNCTION pgroonga.score("row" record)
 	STRICT;
 
 CREATE FUNCTION pgroonga.table_name(indexName cstring)
-	RETURNS cstring
+	RETURNS text
 	AS 'MODULE_PATHNAME', 'pgroonga_table_name'
 	LANGUAGE C
+	STABLE
+	STRICT;
+
+CREATE FUNCTION pgroonga.command(groongaCommand text)
+	RETURNS text
+	AS 'MODULE_PATHNAME', 'pgroonga_command'
+	LANGUAGE C
 	VOLATILE
 	STRICT;
 
-CREATE FUNCTION pgroonga.command(groongaCommand text)
+CREATE FUNCTION pgroonga.command(groongaCommand text, arguments text[])
 	RETURNS text
 	AS 'MODULE_PATHNAME', 'pgroonga_command'
 	LANGUAGE C
@@ -27,30 +34,128 @@ CREATE FUNCTION pgroonga.snippet_html(target text, keywords text[])
 	RETURNS text[]
 	AS 'MODULE_PATHNAME', 'pgroonga_snippet_html'
 	LANGUAGE C
-	VOLATILE
+	IMMUTABLE
 	STRICT;
 
 CREATE FUNCTION pgroonga.highlight_html(target text, keywords text[])
 	RETURNS text
 	AS 'MODULE_PATHNAME', 'pgroonga_highlight_html'
 	LANGUAGE C
-	VOLATILE
+	IMMUTABLE
 	STRICT;
 
 CREATE FUNCTION pgroonga.match_positions_byte(target text, keywords text[])
 	RETURNS integer[2][]
 	AS 'MODULE_PATHNAME', 'pgroonga_match_positions_byte'
 	LANGUAGE C
-	VOLATILE
+	IMMUTABLE
+	STRICT;
+
+CREATE FUNCTION pgroonga.match_positions_character(target text, keywords text[])
+	RETURNS integer[2][]
+	AS 'MODULE_PATHNAME', 'pgroonga_match_positions_character'
+	LANGUAGE C
+	IMMUTABLE
 	STRICT;
 
 CREATE FUNCTION pgroonga.query_extract_keywords(query text)
 	RETURNS text[]
 	AS 'MODULE_PATHNAME', 'pgroonga_query_extract_keywords'
 	LANGUAGE C
+	IMMUTABLE
+	STRICT;
+
+CREATE FUNCTION pgroonga.flush(indexName cstring)
+	RETURNS bool
+	AS 'MODULE_PATHNAME', 'pgroonga_flush'
+	LANGUAGE C
 	VOLATILE
 	STRICT;
 
+CREATE FUNCTION pgroonga.command_escape_value(value text)
+	RETURNS text
+	AS 'MODULE_PATHNAME', 'pgroonga_command_escape_value'
+	LANGUAGE C
+	IMMUTABLE
+	STRICT;
+
+CREATE FUNCTION pgroonga.query_escape(query text)
+	RETURNS text
+	AS 'MODULE_PATHNAME', 'pgroonga_query_escape'
+	LANGUAGE C
+	IMMUTABLE
+	STRICT;
+
+CREATE FUNCTION pgroonga.escape(value text)
+	RETURNS text
+	AS 'MODULE_PATHNAME', 'pgroonga_escape_string'
+	LANGUAGE C
+	IMMUTABLE
+	STRICT;
+
+CREATE FUNCTION pgroonga.escape(value text, special_characters text)
+	RETURNS text
+	AS 'MODULE_PATHNAME', 'pgroonga_escape_string'
+	LANGUAGE C
+	IMMUTABLE
+	STRICT;
+
+CREATE FUNCTION pgroonga.escape(value boolean)
+	RETURNS text
+	AS 'MODULE_PATHNAME', 'pgroonga_escape_boolean'
+	LANGUAGE C
+	IMMUTABLE
+	STRICT;
+
+CREATE FUNCTION pgroonga.escape(value int2)
+	RETURNS text
+	AS 'MODULE_PATHNAME', 'pgroonga_escape_int2'
+	LANGUAGE C
+	IMMUTABLE
+	STRICT;
+
+CREATE FUNCTION pgroonga.escape(value int4)
+	RETURNS text
+	AS 'MODULE_PATHNAME', 'pgroonga_escape_int4'
+	LANGUAGE C
+	IMMUTABLE
+	STRICT;
+
+CREATE FUNCTION pgroonga.escape(value int8)
+	RETURNS text
+	AS 'MODULE_PATHNAME', 'pgroonga_escape_int8'
+	LANGUAGE C
+	IMMUTABLE
+	STRICT;
+
+CREATE FUNCTION pgroonga.escape(value float4)
+	RETURNS text
+	AS 'MODULE_PATHNAME', 'pgroonga_escape_float8'
+	LANGUAGE C
+	IMMUTABLE
+	STRICT;
+
+CREATE FUNCTION pgroonga.escape(value float8)
+	RETURNS text
+	AS 'MODULE_PATHNAME', 'pgroonga_escape_float8'
+	LANGUAGE C
+	IMMUTABLE
+	STRICT;
+
+CREATE FUNCTION pgroonga.escape(value timestamp)
+	RETURNS text
+	AS 'MODULE_PATHNAME', 'pgroonga_escape_timestamptz'
+	LANGUAGE C
+	IMMUTABLE
+	STRICT;
+
+CREATE FUNCTION pgroonga.escape(value timestamptz)
+	RETURNS text
+	AS 'MODULE_PATHNAME', 'pgroonga_escape_timestamptz'
+	LANGUAGE C
+	IMMUTABLE
+	STRICT;
+
 CREATE FUNCTION pgroonga.match_term(target text, term text)
 	RETURNS bool
 	AS 'MODULE_PATHNAME', 'pgroonga_match_term_text'
@@ -224,44 +329,58 @@ CREATE FUNCTION pgroonga.options(internal)
 	AS 'MODULE_PATHNAME', 'pgroonga_options'
 	LANGUAGE C;
 
-DELETE FROM pg_catalog.pg_am WHERE amname = 'pgroonga';
-INSERT INTO pg_catalog.pg_am VALUES(
-	'pgroonga',	-- amname
-	19,		-- amstrategies
-	0,		-- amsupport
-	true,		-- amcanorder
-	true,		-- amcanorderbyop
-	true,		-- amcanbackward
-	true,		-- amcanunique
-	true,		-- amcanmulticol
-	true,		-- amoptionalkey
-	true,		-- amsearcharray
-	false,		-- amsearchnulls
-	false,		-- amstorage
-	true,		-- amclusterable
-	false,		-- ampredlocks
-	0,		-- amkeytype
-	'pgroonga.insert',	-- aminsert
-	'pgroonga.beginscan',	-- ambeginscan
-	'pgroonga.gettuple',	-- amgettuple
-	'pgroonga.getbitmap',	-- amgetbitmap
-	'pgroonga.rescan',	-- amrescan
-	'pgroonga.endscan',	-- amendscan
-	0,		-- ammarkpos,
-	0,		-- amrestrpos,
-	'pgroonga.build',	-- ambuild
-	'pgroonga.buildempty',	-- ambuildempty
-	'pgroonga.bulkdelete',	-- ambulkdelete
-	'pgroonga.vacuumcleanup',	-- amvacuumcleanup
-	'pgroonga.canreturn',		-- amcanreturn
-	'pgroonga.costestimate',	-- amcostestimate
-	'pgroonga.options'	-- amoptions
-);
+DO LANGUAGE plpgsql $$
+BEGIN
+	EXECUTE 'DROP ACCESS METHOD IF EXISTS pgroonga CASCADE';
+	CREATE FUNCTION pgroonga.handler(internal)
+		RETURNS index_am_handler
+		AS 'MODULE_PATHNAME', 'pgroonga_handler'
+		LANGUAGE C;
+	EXECUTE 'CREATE ACCESS METHOD pgroonga ' ||
+		'TYPE INDEX ' ||
+		'HANDLER pgroonga.handler';
+EXCEPTION
+	WHEN syntax_error THEN
+		DELETE FROM pg_am WHERE amname = 'pgroonga';
+		INSERT INTO pg_am VALUES(
+			'pgroonga',	-- amname
+			21,		-- amstrategies
+			0,		-- amsupport
+			true,		-- amcanorder
+			true,		-- amcanorderbyop
+			true,		-- amcanbackward
+			true,		-- amcanunique
+			true,		-- amcanmulticol
+			true,		-- amoptionalkey
+			true,		-- amsearcharray
+			false,		-- amsearchnulls
+			false,		-- amstorage
+			true,		-- amclusterable
+			false,		-- ampredlocks
+			0,		-- amkeytype
+			'pgroonga.insert',	-- aminsert
+			'pgroonga.beginscan',	-- ambeginscan
+			'pgroonga.gettuple',	-- amgettuple
+			'pgroonga.getbitmap',	-- amgetbitmap
+			'pgroonga.rescan',	-- amrescan
+			'pgroonga.endscan',	-- amendscan
+			0,		-- ammarkpos,
+			0,		-- amrestrpos,
+			'pgroonga.build',	-- ambuild
+			'pgroonga.buildempty',	-- ambuildempty
+			'pgroonga.bulkdelete',	-- ambulkdelete
+			'pgroonga.vacuumcleanup',	-- amvacuumcleanup
+			'pgroonga.canreturn',		-- amcanreturn
+			'pgroonga.costestimate',	-- amcostestimate
+			'pgroonga.options'	-- amoptions
+		);
+END;
+$$;
 
 CREATE OPERATOR CLASS pgroonga.text_full_text_search_ops DEFAULT FOR TYPE text
 	USING pgroonga AS
-		OPERATOR 6 pg_catalog.~~,
-		OPERATOR 7 pg_catalog.~~*,
+		OPERATOR 6 ~~,
+		OPERATOR 7 ~~*,
 		OPERATOR 8 %%,
 		OPERATOR 9 @@;
 
@@ -361,8 +480,7 @@ BEGIN
 		FROM pg_type
 		WHERE typname = 'jsonb';
 
-	IF FOUND
-	THEN
+	IF FOUND THEN
 		CREATE FUNCTION pgroonga.match_query(jsonb, text)
 			RETURNS bool
 			AS 'MODULE_PATHNAME', 'pgroonga_match_jsonb'
@@ -386,8 +504,8 @@ $$;
 
 CREATE OPERATOR CLASS pgroonga.text_regexp_ops FOR TYPE text
 	USING pgroonga AS
-		OPERATOR 6 pg_catalog.~~,
-		OPERATOR 7 pg_catalog.~~*,
+		OPERATOR 6 ~~,
+		OPERATOR 7 ~~*,
 		OPERATOR 10 @~;
 
 CREATE OPERATOR CLASS pgroonga.varchar_regexp_ops FOR TYPE varchar
@@ -500,10 +618,36 @@ CREATE OPERATOR &?> (
 	RIGHTARG = text[]
 );
 
+CREATE FUNCTION pgroonga.prefix_contain_text_array(text[], text)
+	RETURNS bool
+	AS 'MODULE_PATHNAME', 'pgroonga_prefix_contain_text_array'
+	LANGUAGE C
+	IMMUTABLE
+	STRICT;
+
+CREATE OPERATOR &^> (
+	PROCEDURE = pgroonga.prefix_contain_text_array,
+	LEFTARG = text[],
+	RIGHTARG = text
+);
+
+CREATE FUNCTION pgroonga.prefix_rk_contain_text_array(text[], text)
+	RETURNS bool
+	AS 'MODULE_PATHNAME', 'pgroonga_prefix_rk_contain_text_array'
+	LANGUAGE C
+	IMMUTABLE
+	STRICT;
+
+CREATE OPERATOR &^~> (
+	PROCEDURE = pgroonga.prefix_rk_contain_text_array,
+	LEFTARG = text[],
+	RIGHTARG = text
+);
+
 CREATE OPERATOR CLASS pgroonga.text_full_text_search_ops_v2 FOR TYPE text
 	USING pgroonga AS
-		OPERATOR 6 pg_catalog.~~,
-		OPERATOR 7 pg_catalog.~~*,
+		OPERATOR 6 ~~,
+		OPERATOR 7 ~~*,
 		OPERATOR 12 &@,
 		OPERATOR 13 &?,
 		OPERATOR 14 &~?,
@@ -515,3 +659,8 @@ CREATE OPERATOR CLASS pgroonga.text_term_search_ops_v2 FOR TYPE text
 	USING pgroonga AS
 		OPERATOR 16 &^,
 		OPERATOR 17 &^~;
+
+CREATE OPERATOR CLASS pgroonga.text_array_term_search_ops_v2 FOR TYPE text[]
+	USING pgroonga AS
+		OPERATOR 20 &^> (text[], text),
+		OPERATOR 21 &^~> (text[], text);
